"""pypyr step to fetch a json file from s3 and put it in context."""
from collections.abc import MutableMapping
import json
import logging
import pypyraws.aws.s3

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def run_step(context):
    """Fetch a json file from s3 and put the json values into context.

    Args:
        - context: pypyr.context.Context. Mandatory. Should contain keys for:
            - s3Fetch: dict. mandatory. Must contain:
                -methodArgs
                    - Bucket: string. s3 bucket name.
                    - Key: string. s3 key name.
                -key. string. If exists, write json structure to this
                               context key. Else json writes to context root.

    All inputs support formatting expressions.

    json parsed from the s3 file will be merged into the
    context. This will overwrite existing values if the same keys are already
    in there. I.e if s3 json has {'eggs' : 'boiled'} and context
    {'eggs': 'fried'} already exists, returned context['eggs'] will be
    'boiled'.
    """
    logger.debug("started")
    fetch_me = pypyraws.aws.s3.get_fetch_input(context, __name__)

    response = pypyraws.aws.s3.get_payload(fetch_me)

    payload = json.load(response)
    logger.debug("successfully parsed json from s3 response bytes")

    destination_key = fetch_me.get('key', None)
    if not destination_key:
        # backward compatibility
        destination_key = fetch_me.get('outKey', None)

    if destination_key:
        logger.debug(f"writing json to context {destination_key}")
        context[destination_key] = payload
    else:
        if not isinstance(payload, MutableMapping):
            raise TypeError(
                'json input should describe an object at the top '
                'level when outKey isn\'t specified. You should have '
                'something like {"key1": "value1", "key2": "value2"} '
                'in the json top-level, not ["value1", "value2"]')

        context.update(payload)

    logger.info("loaded s3 json into pypyr context")

    logger.debug("done")



"""pypyr step to fetch a yaml file from s3 and put it in context."""
from collections.abc import MutableMapping
import logging
import pypyraws.aws.s3
import ruamel.yaml as yaml

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def run_step(context):
    """Fetch a yaml file from s3 and put the yaml values into context.

    Args:
        - context: pypyr.context.Context. Mandatory. Should contain keys for:
            - s3Fetch: dict. mandatory. Must contain:
                - Bucket: string. s3 bucket name.
                - Key: string. s3 key name.
            - key. string. If exists, write yaml structure to this
                           context key. Else yaml writes to context root.

    yaml parsed from the s3 file will be merged into the
    context. This will overwrite existing values if the same keys are already
    in there. I.e if s3 yaml has {'eggs' : 'boiled'} and context
    {'eggs': 'fried'} already exists, returned context['eggs'] will be
    'boiled'.
    """
    logger.debug("started")

    fetch_me = context.get_formatted('s3Fetch')

    response = pypyraws.aws.s3.get_payload(fetch_me)

    yaml_loader = yaml.YAML(typ='safe', pure=True)
    payload = yaml_loader.load(response)
    logger.debug("successfully parsed yaml from s3 response bytes")

    destination_key = fetch_me.get('key', None)
    if not destination_key:
        # backward compatibility
        destination_key = fetch_me.get('outKey', None)

    if destination_key:
        logger.debug(f"writing yaml to context {destination_key}")
        context[destination_key] = payload
    else:
        if not isinstance(payload, MutableMapping):
            raise TypeError(
                "yaml input should describe a dictionary at the top "
                "level when outKey isn't specified. You should have "
                "something like \n'key1: value1'\n key2: value2'\n"
                "in the yaml top-level, not \n'- value1\n - value2'")
        context.update(payload)

    logger.info("loaded s3 yaml into pypyr context")

    logger.debug("done")



"""s3 higher-level functions."""
import logging
import pypyraws.aws.service
from pypyr.errors import KeyNotInContextError

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def get_payload(fetch_me):
    """Get object from s3, reads underlying http stream, returns bytes.

    Args:
        fetch_me (dict): Mandatory. Must contain key:
            - methodArgs
                - Bucket: string. s3 bucket name.
                - Key: string. s3 key name.

    Returns:
        bytes: payload of the s3 obj in bytes

    Raises:
        KeyNotInContextError: s3Fetch or s3Fetch.methodArgs missing
    """
    logger.debug("started")

    try:
        operation_args = fetch_me['methodArgs']
    except KeyError as err:
        raise KeyNotInContextError(
            "s3Fetch missing required key for pypyraws.steps.s3fetch step: "
            "methodArgs") from err

    client_args = fetch_me.get('clientArgs', None)

    response = pypyraws.aws.service.operation_exec(
        service_name='s3',
        method_name='get_object',
        client_args=client_args,
        operation_args=operation_args)

    logger.debug("reading response stream")
    payload = response['Body']
    logger.debug("returning response bytes")

    logger.debug("done")
    return payload


def get_fetch_input(context, caller):
    """Get s3Fetch formatted context.

    Args:
        context (pypyr.context.Context(): Mandatory. Must contain:
            - s3Fetch: dict. mandatory. Must contain:
                - methodArgs
                    - Bucket: string. s3 bucket name.
                    - Key: string. s3 key name.
        caller (string): Mandatory. __name__ of caller.

    Returns:
        Formatted dict under the context['s3Fetch']

    Raises:
       KeyNotInContextError: s3Fetch or s3Fetch.methodArgs missing
       KeyInContextHasNoValueError: s3Fetch doesn't have a value.
    """
    context.assert_key_has_value(key='s3Fetch', caller=caller)
    return context.get_formatted('s3Fetch')



"""pypyr step that runs any boto3 low-level client waiter."""
import logging
import pypyraws.aws.service
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError


# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def run_step(context):
    """Run any low-level boto3 client wait from get_waiter.

    All of the awsWaitIn descendant values support {key}
    string interpolation.

    Args:
        context:
            Dictionary. Mandatory.
            Requires the following context keys in context:
                - awsWaitIn. dict. mandatory. Contains keys:

                  The awsWaitIn dictionary should contain:
                    - serviceName: mandatory. String for service name.
                      Available services here:
                      http://boto3.readthedocs.io/en/latest/reference/services/
                    - waiterName: mandatory. String. Name of waiter.
                    - waiterArgs: optional. Dict. kwargs for get_waiter
                    - waitArgs: optional. Dict. kwargs for wait

    Returns: None

    Raises:
        botocore.exceptions.ClientError: Anything inside boto went wrong.
        botocore.exceptions.WaiterError: General error.
        botocore.exceptions.FailureStateError: Status changed to a different
                                               state
        botocore.exceptions.TooManyAttemptsError: Timed out waiting for desired
                                                  state.
        pypyr.errors.KeyNotInContextError: awsClientIn missing in context.
        pypyr.errors.KeyInContextHasNoValueError: awsClientIn exists but is
                                                None.
    """
    logger.debug("started")
    client_in, service_name, waiter_name = get_waiter_args(context)

    waiter_args = client_in.get('waiterArgs', None)

    wait_args = client_in.get('waitArgs', None)

    logger.info(f"Waiting for {waiter_name} on aws {service_name}.")

    pypyraws.aws.service.waiter(
        service_name=service_name,
        waiter_name=waiter_name,
        waiter_args=waiter_args,
        wait_args=wait_args)

    logger.debug("done")


def get_waiter_args(context):
    """Get required args from context for this step.

    Args:
        context - dict. context.

    Returns:
        tuple(client_in, service_name, waiter_name)

    Raises:
        pypyr.errors.KeyNotInContextError: Required key missing in context.
        pypyr.errors.KeyInContextHasNoValueError: Required key exists but is
                                                  empty or None.
    """
    context.assert_key_has_value(key='awsWaitIn', caller=__name__)
    client_in = context.get_formatted('awsWaitIn')
    try:
        service_name = client_in['serviceName']
        waiter_name = client_in['waiterName']
    except KeyError as err:
        raise KeyNotInContextError(
            "awsWaitIn missing required key for pypyraws.steps.wait: "
            f"{err}"
        ) from err

    # of course, if They went and made it a bool and True this will pass.
    if not (service_name and service_name.strip()):
        raise KeyInContextHasNoValueError(
            'serviceName required in awsWaitIn for pypyraws.steps.wait')

    if not (waiter_name and waiter_name.strip()):
        raise KeyInContextHasNoValueError(
            'waiterName required in awsWaitIn for pypyraws.steps.wait')

    return client_in, service_name, waiter_name



"""Custom exceptions for pypyraws.

All pypyraws specific exceptions derive from pypyr root Error.
"""

from pypyr.errors import PlugInError


class Error(PlugInError):
    """Base class for all pypyraws exceptions."""


class WaitTimeOut(Error):
    """Aws resource that did not finish processing within wait limit."""



"""pypyr step that runs any boto3 low-level client method."""
import logging
import pypyraws.aws.service
import pypyraws.contextargs as contextargs


# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def run_step(context):
    """Execute any low-level boto3 client method.

    All of the awsClientIn descendant values support {key}
    string interpolation.

    Args:
        context:
            Dictionary. Mandatory.
            Requires the following context keys in context:
                - awsClientIn. dict. mandatory. Contains keys:

            The awsClientIn dictionary should contain:
                - serviceName: mandatory. String for service name.
                               Available services here:
                  http://boto3.readthedocs.io/en/latest/reference/services/
                - methodName: mandatory. String. Name of method to execute.
                - clientArgs: optional. Dict. kwargs for the boto client ctor.
                - methodArgs: optional. Dict. kwargs for the client method call

    Returns: None. Although there is no return, this does add awsClientOut to
             context.

             Adds key to context for response from aws client:
                - awsClientOut. Dictionary containing the full aws response.

    Raises:
        botocore.exceptions.ClientError: Anything inside boto went wrong.
        pypyr.errors.KeyNotInContextError: awsClientIn missing in context.
        pypyr.errors.KeyInContextHasNoValueError: awsClientIn exists but is
                                                  None.
    """
    logger.debug("started")
    context.assert_key_has_value('awsClientIn', __name__)
    client_in = context.get_formatted('awsClientIn')

    (service_name,
     method_name,
     client_args,
     method_args) = contextargs.get_awsclient_args(client_in, __name__)

    context['awsClientOut'] = pypyraws.aws.service.operation_exec(
        service_name=service_name,
        method_name=method_name,
        client_args=client_args,
        operation_args=method_args)

    logger.debug("aws response in context['awsClientOut']")
    logger.info(f"Executed {method_name} on aws {service_name}.")

    logger.debug("done")



"""Prepare aws arguments from the pypyr context."""
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError


def get_awsclient_args(input_dict, calling_module_name):
    """Get required args from context for awsClientIn type steps.

    Doesn't do any formatting. You gotta format before you get here.

    Args:
        input_dict (dict): dict-like structure containing awsClientIn key.
        calling_module_name: string. This is just to make a friendly error msg
                             should something go wrong.

    Returns:
        tuple(service_name, method_name, client_args, method_args)

    Raises:
        pypyr.errors.KeyNotInContextError: Required key missing in context.
        pypyr.errors.KeyInContextHasNoValueError: Required key exists but is
                                                  empty or None.
    """
    try:
        service_name = input_dict['serviceName']
        method_name = input_dict['methodName']
    except KeyError as err:
        raise KeyNotInContextError(
            f"awsClientIn missing required key for {calling_module_name}: "
            f"{err}"
        ) from err

    if not (service_name and service_name.strip()):
        raise KeyInContextHasNoValueError(
            f'serviceName required in awsClientIn for {calling_module_name}')

    if not (method_name and method_name.strip()):
        raise KeyInContextHasNoValueError(
            f'methodName required in awsClientIn for {calling_module_name}')

    client_args = input_dict.get('clientArgs', None)
    method_args = input_dict.get('methodArgs', None)

    return service_name, method_name, client_args, method_args



"""pypyr step that prepares context for an ecs waiter."""
import logging
from pypyr.errors import KeyNotInContextError


# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def run_step(context):
    """Run me after an ecs task run or stop to prepare an ecs waiter.

    Prepares the awsWaitIn context key for pypyraws.steps.wait

    Use this step after any of the following methods if you want to use one of
    the ecs waiters to wait for a specific state:
    - describe_services
    - describe_tasks
    - list_services - specify awsEcsWaitPrepCluster if you don't want default
    - list_tasks - specify awsEcsWaitPrepCluster if you don't want default
    - run_task
    - start_task
    - stop_task
    - update_service

    You don't have to use this step, you could always just construct the
    awsWaitIn dictionary in context yourself. It just so happens this step
    saves you some legwork to do so.

    Args:
        context:
            Dictionary. Mandatory.
            Requires the following context keys in context:
                - awsClientOut. dict. mandatory. This is the context key that
                  any ecs command executed by pypyraws.steps.service adds.
                  Chances are pretty good you don't want to construct this by
                  hand yourself - the idea is to use the output as generated by
                  one of the supported ecs methods.
                - awsEcsWaitPrepCluster. string. optional. The short name or
                  full arn of the cluster that hosts the task to describe. If
                  you do not specify a cluster, the default cluster is assumed.
                  For most of the ecs methods the code automatically deduces
                  the cluster from awsClientOut, so don't worry about it.
                  But, when following list_services and list_tasks, you have to
                  specify this parameter. Specifying this parameter will
                  override any automatically deduced cluster arn.

    Returns:
        None.
        Overwrites the awsWaitIn key in context. The new awsWaitIn will contain
        waitArgs filled with the task or service arns found in awsClientOut.

    Raises:
        pypyr.errors.KeyNotInContextError: awsClientOut missing in context.
        pypyr.errors.KeyInContextHasNoValueError: awsClientOut exists but is
                                                  None.
    """
    logger.debug("started")

    context.assert_key_has_value('awsClientOut', __name__)

    if 'awsEcsWaitPrepCluster' in context:
        logger.debug("awsEcsWaitPrepCluster specified in input context")
        # awsEcsWaitPrepCluster always overrides automatically deduced cluster
        cluster = context['awsEcsWaitPrepCluster']
    else:
        cluster = None

    parse_me = context['awsClientOut']
    isTask = False
    isService = False

    if 'service' in parse_me:
        logger.debug("Found 'service' in awsClientOut")
        if cluster is None:
            cluster = parse_me['service']['clusterArn']
        arn_list = [parse_me['service']['serviceArn']]
        isService = True
    elif 'serviceArns' in parse_me:
        logger.debug("Found 'serviceArns' in awsClientOut")
        # no cluster arn is available for these
        arn_list = parse_me['serviceArns']
        isService = True
    elif 'services' in parse_me:
        logger.debug("Found 'services' in awsClientOut")
        if cluster is None:
            # so happens aws methods returning 'services' all do so with only 1
            # cluster input.
            cluster = parse_me['services'][0]['clusterArn']
        arn_list = [svc['serviceArn']
                    for svc in parse_me['services']]
        isService = True
    elif 'task' in parse_me:
        logger.debug("Found 'task' in awsClientOut")
        if cluster is None:
            cluster = parse_me['task']['clusterArn']
        arn_list = [parse_me['task']['taskArn']]
        isTask = True
    elif 'taskArns' in parse_me:
        logger.debug("Found 'taskArns' in awsClientOut")
        # no cluster arn is available for these
        arn_list = parse_me['taskArns']
        isTask = True
    elif 'tasks' in parse_me:
        logger.debug("Found 'tasks' in awsClientOut")
        if cluster is None:
            # so happens the aws methods that returns 'tasks' all do so with
            # only 1 cluster input, i.e cluster will be the same for all of the
            # return tasks
            cluster = parse_me['tasks'][0]['clusterArn']
        arn_list = [task['taskArn'] for task in parse_me['tasks']]
        isTask = True
    else:
        raise KeyNotInContextError('Run ecswaitprep after an ecs method that '
                                   'does something with services or tasks. '
                                   'Couldn\'t find service, serviceArns, '
                                   'services, task, taskArns or tasks in '
                                   'awsClientOut.')

    waiter_dict = {}
    if cluster is None:
        logger.debug("No cluster specified. Waiter will use default cluster.")
    else:
        logger.debug(f"{cluster} cluster specified.")
        waiter_dict['cluster'] = cluster

    if isTask:
        logger.debug("Adding task arns")
        waiter_dict['tasks'] = arn_list

    if isService:
        logger.debug("Adding service arns")
        waiter_dict['services'] = arn_list

    if 'awsWaitIn' not in context:
        context['awsWaitIn'] = {}
    context['awsWaitIn']['waitArgs'] = waiter_dict
    logger.info("added context['awsWaitIn']['waitArgs']")
    logger.debug("done")



"""aws service client.

This works with the boto low-level service client object.
"""
import boto3
import logging

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def operation_exec(service_name,
                   method_name,
                   client_args=None,
                   operation_args=None):
    """Execute operation on boto low-level service client.

    Args:
        service_name: String. Name of service. Available services here:
                      http://boto3.readthedocs.io/en/latest/reference/services/
        method_name: String. Execute this method against the service_name
                     client. Check boto3 docs for available methods for
                     service_name.
        client_args: dict. Passed to the kwargs of the
                     boto3.client(*args, **kwargs) function.
        operation_args: dict. These are passed as kwargs to the method_name
                        when executing it.

    Returns:
        dict. Response from
                    service_name(client_args).method_name(operation_args)
              Be aware this could well be None if the service operation doesn't
              have a return.
              Code safe: use the default 2nd args on dict.get(key, default)
              to specify a default value where there is a chance of the value
              not existing. Don't bother if you're cheerful about KeyError.
    """
    logger.debug("started")
    if client_args is None:
        client = boto3.client(service_name)
        logger.debug(f"boto client instantiated {service_name} with no "
                     "constructor args")
    else:
        client = boto3.client(service_name, **client_args)
        logger.debug(f"boto client instantiated {service_name} with "
                     "constructor args")

    # dynamically executing method_name against the client and passing it
    # operation_args while it's at it.
    response = getattr(client, method_name)(**operation_args)
    logger.debug(f"Executed {method_name} on {service_name}.")
    logger.debug("done")
    return response


def waiter(service_name, waiter_name, waiter_args=None, wait_args=None):
    """Wait for an aws low-level client operation to reach poll state.

    Waiters use a client's service operations to poll the status of an AWS
    resource and suspend execution until the AWS resource reaches the state
    that the waiter is polling for or a failure occurs while polling.

    Default polls every 6 seconds until a successful state.
    Error after 100 polls.

    Args:
        service_name: String. Name of service. Available services here:
                      http://boto3.readthedocs.io/en/latest/reference/services/
        waiter_name: String. Get this waiter from the service_name
                     client. Check boto3 docs for available waiters for
                     service_name.
        waiter_args: dict. Passed to the kwargs of the
                     boto3.get_waiter(*args, **kwargs) function.
        wait_args: dict. These are passed as kwargs to the waiter's wait method
                   get_waiter(waiter_name).wait(**kwargs)

    Returns: None

    Raises:
        botocore.exceptions.WaiterError: General error.
        botocore.exceptions.FailureStateError: Status changed to a different
                                               state
        botocore.exceptions.TooManyAttemptsError: Timed out waiting for desired
                                                  state.
    """
    logger.debug("started")

    client = boto3.client(service_name)

    if waiter_args is None:
        waiter = client.get_waiter(waiter_name)
        logger.debug(f"boto client got waiter {waiter_name} on {service_name} "
                     "with no waiter args")
    else:
        waiter = client.get_waiter(waiter_name, **waiter_args)
        logger.debug(f"boto client got waiter {waiter_name} on {service_name} "
                     "with waiter args")

    waiter.wait(**wait_args)
    logger.debug("done")



"""Version information."""

import platform

__version__ = '1.3.0'


def get_version():
    """Return package-name __version__ python python_version."""
    return (f'pypyraws {__version__} '
            f'python {platform.python_version()}')


if __name__ == '__main__':
    """Entry point for script execution.

    Makes it easy to get version number from cli from outside the package.
    """
    print(__version__)



"""pypyr step that creates a custom waiter for any aws client operation."""
import logging
from pypyr.utils.asserts import assert_key_has_value
from pypyr.utils.poll import wait_until_true
import pypyraws.aws.service
import pypyraws.contextargs as contextargs
from pypyraws.errors import WaitTimeOut

# pypyr logger means the log level will be set correctly and output formatted.
logger = logging.getLogger(__name__)


def run_step(context):
    """Wait for any aws client operation.

    All of the awsWaitFor descendant values support {key}
    string interpolation, except waitForField.

    Args:
        context:
            Dictionary. Mandatory.
            Requires the following context keys in context:
                - awsWaitFor. dict. mandatory. Contains keys:
                    - awsClientIn. dict. mandatory. This is the same as for the
                      pypyraws.steps.client in step. Contains keys:
                      - serviceName: mandatory. String for service name.
                                       Available services here:
                        http://boto3.readthedocs.io/en/latest/reference/services/
                      - methodName: mandatory. String. Name of method to
                                    execute.
                      - clientArgs: optional. Dict. kwargs for the boto client
                                    ctor.
                      - methodArgs: optional. Dict. kwargs for the client
                                    method call
                    - waitForField: mandatory. string. format expression for
                                    field name to check in awsClient response.
                    - toBe: mandatory. string. string. Stop waiting when
                            waitForField equals this value.
                    - pollInterval: optional. int. In seconds. Default to 30.
                    - maxAttempts: optional. int. Default 10.
                    - errorOnWaitTimeout: optional. Default True. Throws error
                                          if maxAttempts exhausted without
                                          reaching toBe value. If false,
                                          step completes without raising
                                          error.

    Returns: None
             Adds key to context:
            - awsWaitForTimedOut: bool. Adds key with value True if
              errorOnWaitTimeout=False and max_attempts exhausted without
              reaching toBe. If steps completes successfully and waitForField's
              value becomes toBe, awsWaitForTimedOut == False.

    Raises:
        pypyr.errors.KeyNotInContextError: awsWaitFor missing in context.
        pypyr.errors.KeyInContextHasNoValueError: awsWaitFor exists but is
                                                None.
        pypyraws.errors.WaitTimeOut: maxAttempts exceeded without waitForField
                                  changing to toBe.
    """
    logger.debug("started")
    context.assert_key_has_value('awsWaitFor', __name__)
    wait_for = context['awsWaitFor']

    assert_key_has_value(wait_for, 'awsClientIn', __name__, 'awsWaitFor')
    aws_client_in = context.get_formatted_value(wait_for['awsClientIn'])

    (service_name,
     method_name,
     client_args,
     method_args) = contextargs.get_awsclient_args(aws_client_in, __name__)

    (wait_for_field,
     to_be,
     poll_interval,
     max_attempts,
     error_on_wait_timeout) = get_poll_args(wait_for, context)

    wait_response = wait_until_true(interval=poll_interval,
                                    max_attempts=max_attempts)(
        execute_aws_client_method)(
        service_name=service_name,
        method_name=method_name,
        client_args=client_args,
        method_args=method_args,
        wait_for_field=wait_for_field,
        to_be=to_be
    )

    if wait_response:
        context['awsWaitForTimedOut'] = False
        logger.info(f"aws {service_name} {method_name} returned {to_be}. "
                    "Pipeline will now continue.")
    else:
        if error_on_wait_timeout:
            context['awsWaitForTimedOut'] = True
            logger.error(f"aws {service_name} {method_name} did not return "
                         f"{to_be} within {max_attempts}. errorOnWaitTimeout "
                         "is True, throwing error")
            raise WaitTimeOut(f"aws {service_name} {method_name} did not "
                              f"return {to_be} within {max_attempts} retries.")
        else:
            context['awsWaitForTimedOut'] = True
            logger.warning(
                f"aws {service_name} {method_name} did NOT return "
                f" {to_be}. errorOnWaitTimeout is False, so pipeline "
                "will proceed to the next step anyway.")

    logger.debug("done")


def execute_aws_client_method(service_name,
                              method_name,
                              client_args,
                              method_args,
                              wait_for_field,
                              to_be):
    """Execute method_name on service_name.

    Args:
        service_name: string. Name of aws service.
        method_name: method to execute.
        client_args: aws client constructor args.
        method_args: method args
        wait_for_field: look for this field in the aws response
        to_be: return True if wait_for_field's value equals this.

    Return:
        True if value of wait_for_field == to_be, False if not.
    """
    logger.debug("started")
    response = pypyraws.aws.service.operation_exec(
        service_name=service_name,
        method_name=method_name,
        client_args=client_args,
        operation_args=method_args)

    wait_for_this_value = wait_for_field.format(**response)
    logger.info(f"{wait_for_field} in aws response is: {wait_for_this_value}")
    if wait_for_this_value == str(to_be):
        logger.debug("Required status reached. The wait is so over.")
        logger.debug("done")
        return True
    else:
        logger.debug("Required status not reached, keep waiting.")
        logger.debug("done")
        return False


def get_poll_args(waitfor_dict, context):
    """Get polling arguments from waitfor_dict.

    Args:
        waitfor_dict: The awsWaitFor dict
        context: the pypyr context

    Returns:
    tuple(wait_for_field,
          to_be,
          poll_interval,
          max_attempts,
          error_on_wait_timeout)
    """
    logger.debug("started")
    wait_for_field = waitfor_dict['waitForField']

    to_be = context.get_formatted_value(waitfor_dict['toBe'])

    poll_interval = context.get_formatted_as_type(
        waitfor_dict.get('pollInterval', 30),
        out_type=float)

    max_attempts = context.get_formatted_as_type(
        waitfor_dict.get('maxAttempts', 10),
        out_type=int)

    error_on_wait_timeout = context.get_formatted_as_type(
        waitfor_dict.get('errorOnWaitTimeout', True),
        out_type=bool)

    logger.debug("done")

    return (wait_for_field,
            to_be,
            poll_interval,
            max_attempts,
            error_on_wait_timeout)
